import 'dart:convert';

import 'package:dart_aas/dart_aas.dart';
import 'package:flutter/foundation.dart';
import 'package:s3i_flutter/src/auth/authentication_manager.dart';
import 'package:s3i_flutter/src/broker/messages/attribute_value_messages.dart';
import 'package:s3i_flutter/src/broker/messages/frame.dart';
import 'package:s3i_flutter/src/broker/messages/i40_message.dart';
import 'package:s3i_flutter/src/broker/messages/message.dart';
import 'package:s3i_flutter/src/broker/messages/service_messages.dart';
import 'package:s3i_flutter/src/broker/messages/user_message.dart';
import 'package:s3i_flutter/src/exceptions/invalid_json_schema_exception.dart';
import 'package:s3i_flutter/src/exceptions/json_missing_key_exception.dart';
import 'package:s3i_flutter/src/exceptions/parse_exception.dart';
import 'package:s3i_flutter/src/utils/json_keys.dart';

import '../../s3i_flutter.dart';

/// The baseclass for all communication interfaces with the `S3I-Broker`.
///
/// There are two different approaches to receive messages from the Broker:
/// - [ActiveBrokerInterface] for interfaces that inform you whenever a new
/// message is available.
/// - [PassiveBrokerInterface] for interfaces where you need to explicitly ask
/// if there are new messages.
abstract class BrokerInterface {
  /// Creates a [BrokerInterface] with the given [authManager].
  BrokerInterface(this.authManager);

  /// The authentication manager used by this instance to get
  /// valid access tokens.
  final AuthenticationManager authManager;

  /// Publishes the given [message] to all [endpoints].
  void sendI40Message(I40Message message, Set<Endpoint> endpoints);

  static I40Message transformJsonToI40Message(Map<String, dynamic> json) {
    final Frame frame = json.containsKey(I40MessageKeys.frame)
        ? Frame.fromJson(json[I40MessageKeys.frame] as Map<String, dynamic>)
        : throw JsonMissingKeyException(I40MessageKeys.frame, json.toString());

    final Reference semanticProtocol = frame.semanticProtocol;
    String semanticProtocolValue = "";
    if (semanticProtocol.keys.isNotEmpty) {
      semanticProtocolValue = semanticProtocol.keys.first.value;
    }
    switch (frame.messageType) {
      case I40MessageType.unkown:
        throw InvalidJsonSchemaException('unknown/unsupported message type', jsonEncode(json));
      case I40MessageType.request:
        switch (semanticProtocolValue) {
          case I40MessageGlobalReferenceValues.invokeOperationAsync:
            return I40InvokeOperationRequest.fromJson(json);
          case I40MessageGlobalReferenceValues.invokeOperationSync:
            return I40InvokeOperationRequest.fromJson(json);
          case I40MessageGlobalReferenceValues.getSubmodelElementByPath:
            return I40GetSubmodelElementByPathRequest.fromJson(json);
          default:
            throw InvalidJsonSchemaException('unknown/unsupported message type', jsonEncode(json));
        }
      case I40MessageType.reply:
        switch (semanticProtocolValue) {
          case I40MessageGlobalReferenceValues.invokeOperationAsync:
            return I40InvokeOperationReply.fromJson(json);
          case I40MessageGlobalReferenceValues.invokeOperationSync:
            return I40InvokeOperationReply.fromJson(json);
          case I40MessageGlobalReferenceValues.getSubmodelElementByPath:
            return I40GetSubmodelElementByPathReply.fromJson(json);
          default:
            throw InvalidJsonSchemaException('unknown/unsupported message type', jsonEncode(json));
        }
      case I40MessageType.event:
        return I40EventMessage.fromJson(json);
    }
  }
}

/// The [PassiveBrokerInterface] is the baseclass for all broker interfaces
/// where the user needs to call a method to see if a new message is available.
abstract class PassiveBrokerInterface extends BrokerInterface {
  /// Creates a new [PassiveBrokerInterface] with the given [authManager].
  PassiveBrokerInterface(AuthenticationManager authManager) : super(authManager);

  /// Returns a [Message] if it's available at the [endpoint], null otherwise.
  Message? getMessage(String endpoint);
}

/// The [ActiveBrokerInterface] is the baseclass for all broker interfaces
/// which inform the user asynchronous (via a callback) when a new message is
/// available.
abstract class ActiveBrokerInterface extends BrokerInterface {
  /// Creates a new [ActiveBrokerInterface] with the given [authManager].
  ActiveBrokerInterface(AuthenticationManager authManager) : super(authManager);

  /// All registered functions which are invoked if an error occurred during the
  /// message consuming.
  final Set<Function(String, Exception)> _callbacksConsumingFailed = <Function(String, Exception)>{};

  /// All registered functions which are invoked if a message could not be sent.
  final Set<Function(Message, Exception)> _callbacksSendMessageFailed = <Function(Message, Exception)>{};

  /// All registered functions which are invoked if a message could not be sent.
  final Set<Function(I40Message, Exception)> _callbacksSendI40MessageFailed = <Function(I40Message, Exception)>{};

  /// All registered functions which are invoked if a message is sent.
  final Set<Function(Message)> _callbacksSendMessageSucceeded = <Function(Message)>{};

  /// All registered functions which are invoked if a message is sent.
  final Set<Function(I40Message)> _callbacksSendI40MessageSucceeded = <Function(I40Message)>{};

  /// All registered functions which are invoked if an invalid message is
  /// received.
  final Set<Function(String, Exception)> _callbacksForInvalidMessage = <Function(String, Exception)>{};

  /// All registered functions which are invoked if an invalid I40 message is
  /// received.
  final Set<Function(String, Exception)> _callbacksForInvalidI40Message = <Function(String, Exception)>{};

  /// All registered functions which are invoked if an [UserMessage] is
  /// received.
  final Set<Function(UserMessage)> _callbacksForUserMessage = <Function(UserMessage)>{};

  /// All registered functions which are invoked if an [I40EventMessage] is
  /// received.
  final Set<Function(I40EventMessage)> _callbacksForI40EventMessage = <Function(I40EventMessage)>{};

  /// All registered functions which are invoked if a [ServiceRequest] is
  /// received.
  final Set<Function(ServiceRequest)> _callbacksForServiceRequest = <Function(ServiceRequest)>{};

  /// All registered functions which are invoked if an [I40InvokeOperationAsyncRequest] is
  /// received.
  final Set<Function(I40InvokeOperationAsyncRequest)> _callbacksForI40InvokeOperationAsyncRequest =
      <Function(I40InvokeOperationAsyncRequest)>{};

  /// All registered functions which are invoked if a [ServiceReply] is
  /// received.
  final Set<Function(ServiceReply)> _callbacksForServiceReply = <Function(ServiceReply)>{};

  /// All registered functions which are invoked if an [I40InvokeOperationAsyncReply] is
  /// received.
  final Set<Function(I40InvokeOperationAsyncReply)> _callbacksForI40InvokeOperationAsyncReply =
      <Function(I40InvokeOperationAsyncReply)>{};

  /// All registered functions which are invoked if a [GetValueRequest] is
  /// received.
  final Set<Function(GetValueRequest)> _callbacksForGetValueRequest = <Function(GetValueRequest)>{};

  /// All registered functions which are invoked if an [I40GetSubmodelElementByPathRequest] is
  /// received.
  final Set<Function(I40GetSubmodelElementByPathRequest)> _callbacksForI40GetSubmodelElementByPathRequest =
      <Function(I40GetSubmodelElementByPathRequest)>{};

  /// All registered functions which are invoked if a [GetValueReply] is
  /// received.
  final Set<Function(GetValueReply)> _callbacksForGetValueReply = <Function(GetValueReply)>{};

  /// All registered functions which are invoked if an [I40GetSubmodelElementByPathReply] is
  /// received.
  final Set<Function(I40GetSubmodelElementByPathReply)> _callbacksForI40GetSubmodelElementByPathReply =
      <Function(I40GetSubmodelElementByPathReply)>{};

  /// External function invoked when token expires in EventBrokerConnector,
  /// allowing to create Binding between Event Queue and Topic Exchange
  late Function() _recreateTopicBinding;

  // TODO(poq): add the other message types

  /// Starts consuming on the endpoint.
  ///
  /// The endpoints should be in the correct format (`s3ib://s3i:`+ UUIDv4
  /// for decrypted communication or `s3ibs://s3i:` + UUIDv4 for encrypted
  /// messages).
  Future<void> startConsuming(String endpoint);

  /// Stops consuming on the endpoint.
  ///
  /// The endpoints should be in the correct format (`s3ib://s3i:`+ UUIDv4
  /// for decrypted communication or `s3ibs://s3i:` + UUIDv4 for encrypted
  /// messages).
  void stopConsuming(String endpoint);

  /// Actively reconnect to Broker, when already connected before
  /// This could be useful when the connection is lost due to background activity
  Future<void> reconnectToBroker();

  /// Check handshakeStatus of the broker with the Server.
  bool isConnected();

  /// PROTECTED: DO NOT USE THIS METHOD UNLESS YOU ARE AN
  /// [ActiveBrokerInterface] !
  ///
  /// This method could be used from an implementation of this interface when
  /// it receives a message. It decodes the [messageString], creates a new
  /// message and notifies the subscribed callbacks.
  ///
  /// This shouldn't throw anything for safe usage.
  void newI40MessageReceived(String messageString) {
    try {
      final Map<String, dynamic> decodedMessage = jsonDecode(messageString) as Map<String, dynamic>;
      final I40Message message = BrokerInterface.transformJsonToI40Message(decodedMessage);

      // if (kDebugMode) {
      //   print('I40Message received: ${message.runtimeType}');
      // }
      if (message is I40EventMessage) {
        _notifyI40EventMessageReceived(message);
      } else if (message is I40InvokeOperationAsyncRequest) {
        _notifyInvokeOperationAsyncRequestReceived(message);
      } else if (message is I40InvokeOperationAsyncReply) {
        _notifyInvokeOperationAsyncReplyReceived(message);
      } else if (message is I40GetSubmodelElementByPathRequest) {
        _notifyGetSubmodelElementByPathRequestReceived(message);
      } else if (message is I40GetSubmodelElementByPathReply) {
        _notifyGetSubmodelElementByPathReplyReceived(message);
      } else {
        _notifyInvalidI40MessageReceived(messageString, ParseException('unknown message type'));
      }
    } on TypeError catch (e) {
      _notifyInvalidI40MessageReceived(messageString, ParseException(e.stackTrace.toString()));
    } on InvalidJsonSchemaException catch (e) {
      _notifyInvalidI40MessageReceived(messageString, e);
    }
  }

  /// Subscribes to all errors during consuming from the endpoints.
  /// (ConsumingFailed-Event).
  ///
  /// The first passed String is the endpoint and the second one is the
  /// exception during the connection process.
  void subscribeConsumingFailed(Function(String, Exception) callback) {
    _callbacksConsumingFailed.add(callback);
  }

  /// Unsubscribes the callback from the ConsumingFailed-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while
  /// subscribing.
  void unsubscribeConsumingFailed(Function(String, Exception) callback) {
    _callbacksConsumingFailed.remove(callback);
  }

  /// Subscribes to all messages which couldn't be sent correctly
  /// (SendMessageFailed-Event).
  ///
  /// The first passed Message is the original message and the second one is
  /// the exception during the sending process.
  void subscribeSendMessageFailed(Function(Message, Exception) callback) {
    _callbacksSendMessageFailed.add(callback);
  }

  /// Unsubscribes the callback from the SendMessageFailed-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while
  /// subscribing.
  void unsubscribeSendMessageFailed(Function(Message, Exception) callback) {
    _callbacksSendMessageFailed.remove(callback);
  }

  /// Subscribes to all messages which couldn't be sent correctly
  /// (SendI40MessageFailed-Event).
  void subscribeSendI40MessageFailed(Function(I40Message, Exception) callback) {
    _callbacksSendI40MessageFailed.add(callback);
  }

  /// Unsubscribes the callback from the SendI40MessageFailed-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while subscribing.
  void unsubscribeSendI40MessageFailed(Function(I40Message, Exception) callback) {
    _callbacksSendI40MessageFailed.remove(callback);
  }

  /// Subscribes to all received messages
  /// (SendMessageSucceeded-Event).
  void subscribeSendMessageSucceeded(Function(Message) callback) {
    _callbacksSendMessageSucceeded.add(callback);
  }

  /// Unsubscribes the callback from the SendMessageSucceeded-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while
  /// subscribing.
  void unsubscribeSendMessageSucceeded(Function(Message) callback) {
    _callbacksSendMessageSucceeded.remove(callback);
  }

  /// Subscribes to all successfully sent I40 messages (SendI40MessageSucceeded-Event).
  void subscribeSendI40MessageSucceeded(Function(I40Message) callback) {
    _callbacksSendI40MessageSucceeded.add(callback);
  }

  /// Unsubscribes the callback from the SendI40MessageSucceeded-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while subscribing.
  void unsubscribeSendI40MessageSucceeded(Function(I40Message) callback) {
    _callbacksSendI40MessageSucceeded.remove(callback);
  }

  /// Subscribes to all received messages which can't be parsed correctly
  /// (InvalidMessageReceived-Event).
  ///
  /// The first passed String is the original message and the second one is the
  /// exception during the parsing process.
  void subscribeInvalidMessageReceived(Function(String, Exception) callback) {
    _callbacksForInvalidMessage.add(callback);
  }

  /// Unsubscribes the callback from the InvalidMessageReceived-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while
  /// subscribing.
  void unsubscribeInvalidMessageReceived(Function(String, Exception) callback) {
    _callbacksForInvalidMessage.remove(callback);
  }

  /// Subscribes to all received I40 messages which can't be parsed correctly
  /// (InvalidI40MessageReceived-Event).
  void subscribeInvalidI40MessageReceived(Function(String, Exception) callback) {
    _callbacksForInvalidI40Message.add(callback);
  }

  /// Unsubscribes the callback from the InvalidI40MessageReceived-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while
  /// subscribing.
  void unsubscribeInvalidI40MessageReceived(Function(String, Exception) callback) {
    _callbacksForInvalidI40Message.remove(callback);
  }

  /// Subscribes to all received user messages (UserMessageReceived-Event).
  void subscribeUserMessageReceived(Function(UserMessage) callback) {
    _callbacksForUserMessage.add(callback);
  }

  /// Unsubscribes the callback from the UserMessageReceived-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while
  /// subscribing.
  void unsubscribeUserMessageReceived(Function(UserMessage) callback) {
    _callbacksForUserMessage.remove(callback);
  }



  /// Subscribes to all received I40 event messages (I40EventMessageReceived-Event).
  void subscribeI40EventMessageReceived(Function(I40EventMessage) callback) {
    _callbacksForI40EventMessage.add(callback);
  }

  /// Unsubscribes the callback from the I40EventMessageReceived-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while subscribing.
  void unsubscribeI40EventMessageReceived(Function(I40EventMessage) callback) {
    _callbacksForI40EventMessage.remove(callback);
  }

  /// Subscribes to all ServiceRequestReceived-Events.
  void subscribeServiceRequestReceived(Function(ServiceRequest) callback) {
    _callbacksForServiceRequest.add(callback);
  }

  /// Unsubscribes the callback from the ServiceRequestReceived-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while
  /// subscribing.
  void unsubscribeServiceRequestReceived(Function(ServiceRequest) callback) {
    _callbacksForServiceRequest.remove(callback);
  }

  /// Subscribes to all I40InvokeOperationAsyncRequest-Events.
  void subscribeI40InvokeOperationAsyncRequestReceived(Function(I40InvokeOperationAsyncRequest) callback) {
    _callbacksForI40InvokeOperationAsyncRequest.add(callback);
  }

  /// Unsubscribes the callback from the I40InvokeOperationAsyncRequest-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while subscribing.
  void unsubscribeI40InvokeOperationAsyncRequestReceived(Function(I40InvokeOperationAsyncRequest) callback) {
    _callbacksForI40InvokeOperationAsyncRequest.remove(callback);
  }

  /// Subscribes to all ServiceReplyReceived-Events.
  void subscribeServiceReplyReceived(Function(ServiceReply) callback) {
    _callbacksForServiceReply.add(callback);
  }

  /// Unsubscribes the callback from the ServiceReplyReceived-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while
  /// subscribing.
  void unsubscribeServiceReplyReceived(Function(ServiceReply) callback) {
    _callbacksForServiceReply.remove(callback);
  }

  /// Subscribes to all I40InvokeOperationAsyncReply-Events.
  void subscribeI40InvokeOperationAsyncReplyReceived(Function(I40InvokeOperationAsyncReply) callback) {
    _callbacksForI40InvokeOperationAsyncReply.add(callback);
  }

  /// Unsubscribes the callback from the I40InvokeOperationAsyncReply-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while subscribing.
  void unsubscribeI40InvokeOperationAsyncReplyReceived(Function(I40InvokeOperationAsyncReply) callback) {
    _callbacksForI40InvokeOperationAsyncReply.remove(callback);
  }

  /// Subscribes to all GetValueRequestReceived-Events.
  void subscribeGetValueRequestReceived(Function(GetValueRequest) callback) {
    _callbacksForGetValueRequest.add(callback);
  }

  /// Unsubscribes the callback from the GetValueRequestReceived-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while
  /// subscribing.
  void unsubscribeGetValueRequestReceived(Function(GetValueRequest) callback) {
    _callbacksForGetValueRequest.remove(callback);
  }

  /// Subscribes to all I40GetSubmodelElementByPathRequest-Events.
  void subscribeI40GetSubmodelElementByPathRequestReceived(Function(I40GetSubmodelElementByPathRequest) callback) {
    _callbacksForI40GetSubmodelElementByPathRequest.add(callback);
  }

  /// Unsubscribes the callback from the I40GetSubmodelElementByPathRequest-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while subscribing.
  void unsubscribeI40GetSubmodelElementByPathRequestReceived(Function(I40GetSubmodelElementByPathRequest) callback) {
    _callbacksForI40GetSubmodelElementByPathRequest.remove(callback);
  }

  /// Subscribes to all GetValueReplyReceived-Events.
  void subscribeGetValueReplyReceived(Function(GetValueReply) callback) {
    _callbacksForGetValueReply.add(callback);
  }

  /// Unsubscribes the callback from the GetValueReplyReceived-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while
  /// subscribing.
  void unsubscribeGetValueReplyReceived(Function(GetValueReply) callback) {
    _callbacksForGetValueReply.remove(callback);
  }

  /// Subscribes to all I40GetSubmodelElementByPathReply-Events.
  void subscribeI40GetSubmodelElementByPathReplyReceived(Function(I40GetSubmodelElementByPathReply) callback) {
    _callbacksForI40GetSubmodelElementByPathReply.add(callback);
  }

  /// Unsubscribes the callback from the I40GetSubmodelElementByPathReply-Event.
  ///
  /// The [callback] needs to be exactly the same as the one used while subscribing.
  void unsubscribeI40GetSubmodelElementByPathReplyReceived(Function(I40GetSubmodelElementByPathReply) callback) {
    _callbacksForI40GetSubmodelElementByPathReply.remove(callback);
  }

  /// PROTECTED: DO NOT USE UNLESS YOU ARE AN [ActiveBrokerInterface].
  void notifyConsumingFailed(String endpoint, Exception error) {
    for (final Function(String, Exception) callback in _callbacksConsumingFailed) {
      callback(endpoint, error);
    }
  }

  /// PROTECTED: DO NOT USE UNLESS YOU ARE AN [ActiveBrokerInterface].
  void notifySendMessageFailed(Message message, Exception error) {
    for (final Function(Message, Exception) callback in _callbacksSendMessageFailed) {
      callback(message, error);
    }
  }

  /// PROTECTED: DO NOT USE UNLESS YOU ARE AN [ActiveBrokerInterface].
  void notifySendI40MessageFailed(I40Message message, Exception error) {
    for (final Function(I40Message, Exception) callback in _callbacksSendI40MessageFailed) {
      callback(message, error);
    }
  }

  /// PROTECTED: DO NOT USE UNLESS YOU ARE AN [ActiveBrokerInterface].
  void notifySendMessageSucceeded(Message message) {
    for (final Function(Message) callback in _callbacksSendMessageSucceeded) {
      callback(message);
    }
  }

  /// PROTECTED: DO NOT USE UNLESS YOU ARE AN [ActiveBrokerInterface].
  void notifySendI40MessageSucceeded(I40Message message) {
    for (final Function(I40Message) callback in _callbacksSendI40MessageSucceeded) {
      callback(message);
    }
  }

  void _notifyInvalidMessageReceived(String original, Exception error) {
    for (final Function(String, Exception) callback in _callbacksForInvalidMessage) {
      callback(original, error);
    }
  }

  void _notifyInvalidI40MessageReceived(String original, Exception error) {
    for (final Function(String, Exception) callback in _callbacksForInvalidI40Message) {
      callback(original, error);
    }
  }

  void _notifyUserMessageReceived(UserMessage message) {
    for (final Function(UserMessage) callback in _callbacksForUserMessage) {
      callback(message);
    }
  }

  void _notifyI40EventMessageReceived(I40EventMessage message) {
    for (final Function(I40EventMessage) callback in _callbacksForI40EventMessage) {
      callback(message);
    }
  }

  void _notifyServiceRequestReceived(ServiceRequest message) {
    for (final Function(ServiceRequest) callback in _callbacksForServiceRequest) {
      callback(message);
    }
  }

  void _notifyInvokeOperationAsyncRequestReceived(I40InvokeOperationAsyncRequest message) {
    for (final Function(I40InvokeOperationAsyncRequest) callback in _callbacksForI40InvokeOperationAsyncRequest) {
      callback(message);
    }
  }

  void _notifyServiceReplyReceived(ServiceReply message) {
    for (final Function(ServiceReply) callback in _callbacksForServiceReply) {
      callback(message);
    }
  }

  void _notifyInvokeOperationAsyncReplyReceived(I40InvokeOperationAsyncReply message) {
    for (final Function(I40InvokeOperationAsyncReply) callback in _callbacksForI40InvokeOperationAsyncReply) {
      callback(message);
    }
  }

  void _notifyGetValueRequestReceived(GetValueRequest message) {
    for (final Function(GetValueRequest) callback in _callbacksForGetValueRequest) {
      callback(message);
    }
  }

  void _notifyGetSubmodelElementByPathRequestReceived(I40GetSubmodelElementByPathRequest message) {
    for (final Function(I40GetSubmodelElementByPathRequest) callback
        in _callbacksForI40GetSubmodelElementByPathRequest) {
      callback(message);
    }
  }

  void _notifyGetValueReplyReceived(GetValueReply message) {
    for (final Function(GetValueReply) callback in _callbacksForGetValueReply) {
      callback(message);
    }
  }

  void _notifyGetSubmodelElementByPathReplyReceived(I40GetSubmodelElementByPathReply message) {
    for (final Function(I40GetSubmodelElementByPathReply) callback in _callbacksForI40GetSubmodelElementByPathReply) {
      callback(message);
    }
  }

  /// Sets function for Topic Binding Recovery in [EventSystemConnector]
  void setTopicBindingRecovery(Function() recFunc) {
    _recreateTopicBinding = recFunc;
  }

  /// Allows to call external Topic Binding Recovery function in BrokerConnector
  /// , e.g. in [BrokerAmqpConnector]
  Future<void> recreateTopicBinding() async {
    await _recreateTopicBinding();
  }
}
