import 'dart:convert';

import 'package:mqtt_client/mqtt_browser_client.dart';
import 'package:mqtt_client/mqtt_client.dart';
import 'package:s3i_flutter/s3i_flutter.dart';

///Creates a new [BrokerMQTTConnector].
///
/// The instance is created with the [authManager] and the optional constructor
/// arguments in the [args] (the string-keys should match the name of the
/// parameters).
///
/// This is needed to use the same interface on web and other platforms.
///
/// The [args] have defaults for the normal message communication in the S3I:
/// - brokerHost = 'rabbitmq.s3i.vswf.dev',
/// - port = 5671,
/// - virtualHost = 's3i',
/// - maxConnectionAttempts = 3,
/// - reconnectWaitTime = const Duration(milliseconds: 1500),
/// - prefetchCount = -1,
/// - exchangeName = 'demo.direct',
/// - exchangeType = ExchangeType.DIRECT
/// onBadCertificate returns always true
ActiveBrokerInterface getActiveBrokerDefaultConnector(AuthenticationManager authManager,
    {Map<String, dynamic> args = const <String, dynamic>{}}) {
  final String brokerHost = args['brokerHost'] as String? ?? 'rabbitmq.s3i.vswf.dev';
  final int port = args['port'] as int? ?? 15675; // Default WebSocket port for MQTT
  final String clientId = args['clientId'] as String? ?? authManager.clientIdentity.id; //should be unique per connection to mqtt
  final String virtualHost = args['virtualHost'] as String? ?? 's3i'; //Doesnt matter as exchange is set in the config of mqtt plugin of rabbitmq, if multiple vhosts are necessary, different ports can be configured
  final String exchangeName =
      args['exchangeName'] as String? ?? 'mqttExchange'; //Doesnt matter as exchange is set in the config of mqtt plugin of rabbitmq

  return BrokerMqttConnector(
    authManager,
    brokerHost: brokerHost,
    port: port,
    clientId: clientId,
    virtualHost: virtualHost,
    exchangeName: exchangeName,
  );
}

/// Creates a new [BrokerMQTTConnector] configured to be used for the
/// Event System.
///
/// See [getActiveBrokerDefaultConnector] for more information.
/// Returns an instance of the BrokerMQTTConnector configured for event handling.
ActiveBrokerInterface getActiveBrokerEventConnector(AuthenticationManager authManager,
    {Map<String, dynamic> args = const <String, dynamic>{}}) {
  args['clientId'] = '${authManager.clientIdentity.id}/event';
  args['exchangeName'] = 'eventExchange';
  return getActiveBrokerDefaultConnector(authManager, args: args);
}

/// This [ActiveBrokerInterface] implementation uses MQTT over websockets to connect on the web
///
/// For more information about RabbitMQ over MQTT see
/// https://www.rabbitmq.com/docs/mqtt
/// https://www.rabbitmq.com/docs/web-mqtt
/// or refer to the documentation of the used
/// [mqtt_client package](https://pub.dev/packages/mqtt_client).
class BrokerMqttConnector extends ActiveBrokerInterface {
  /// MQTT server hostname
  final String brokerHost;

  /// MQTT WebSocket port
  final int port;

  /// Client identifier
  /// should be unique per connection to mqtt
  final String clientId;

  /// Virtual host for the MQTT broker
  /// currently not used, but if necessary different ports per vhost can be configured in rabbitmq mqtt plugin and then mapped
  final String virtualHost;

  /// Exchange name for MQTT topics, not used, exchangeName is configured in the rabbitmq mqtt plugin
  final String exchangeName;

  /// MQTT client instance
  late MqttClient _mqttClient;

  /// Store connected endpoints
  final Set<String> _connectedEndpoints = {};

  BrokerMqttConnector(AuthenticationManager authManager,
      {required this.brokerHost,
      required this.port,
      required this.clientId,
      required this.virtualHost,
      required this.exchangeName})
      : super(authManager) {
    _mqttClient = MqttBrowserClient.withPort('ws://$brokerHost/ws', clientId, port);
    _mqttClient.websocketProtocols = ['mqtt'];
    _mqttClient.logging(on: true);
    // _mqttClient.setProtocolV311();
    _mqttClient.keepAlivePeriod = 20;

    // Set up connection callback
    _mqttClient.onConnected = () {
      print('MQTT Connected');
      _reconnectEndpoints();
    };

    _mqttClient.onDisconnected = () {
      print('MQTT Disconnected');
    };

    _mqttClient.onSubscribed = (String topic) {
      print('Subscribed to $topic');
    };

    _mqttClient.onSubscribeFail = (String topic) {
      print('Failed to subscribe to $topic');
    };
  }

  /// Connect to the MQTT broker
  Future<String> connectToBroker() async {
    if (_mqttClient.connectionStatus!.state == MqttConnectionState.connected) return 'already connected';

    try {
      await _mqttClient.connect(
          'guest', 'guest'); // TODO guest/guest is only valid on localhost, replace with real credentials

      if (_mqttClient.connectionStatus!.state == MqttConnectionState.connected) {
        print('Successfully connected to MQTT broker');
        return '';
      } else {
        print('Connection failed: ${_mqttClient.connectionStatus}');
        return 'MQTT connection failed';
      }
    } catch (e) {
      print('MQTT connection exception: $e');
      _mqttClient.disconnect();
      return 'Connection exception: $e';
    }
  }

  /// Disconnect from the MQTT broker
  Future<void> disconnectFromBroker() async {
    _connectedEndpoints.clear();
    _mqttClient.disconnect();
    print('Disconnected from MQTT broker');
  }

  /// Publish a message to the topic
  @override
  Future<void> sendI40Message(I40Message message, Set<Endpoint> endpoints) async {
    if (_mqttClient.connectionStatus!.state != MqttConnectionState.connected) {
      final String creationInfo = await connectToBroker();
      if (creationInfo.isNotEmpty) {
        notifySendI40MessageFailed(message, S3IException(creationInfo));
        return;
      }
    }

    for (final Endpoint endpoint in endpoints) {
      final builder = MqttClientPayloadBuilder();
      builder.addString(jsonEncode(message.toJson()));

      try {
        _mqttClient.publishMessage(endpoint.endpoint, MqttQos.atMostOnce, builder.payload!);
        notifySendI40MessageSucceeded(message);
        print('Message sent to ${endpoint.endpoint}');
      } catch (e) {
        notifySendI40MessageFailed(message, S3IException('Failed to publish message: $e'));
      }
    }
  }

  /// Subscribe to a topic
  @override
  Future<void> startConsuming(String endpoint) async {
    if (_mqttClient.connectionStatus!.state != MqttConnectionState.connected) {
      final String creationInfo = await connectToBroker();
      if (creationInfo.isNotEmpty) {
        notifyConsumingFailed(endpoint, S3IException(creationInfo));
        return;
      }
    }

    try {
      _mqttClient.subscribe(endpoint, MqttQos.atMostOnce);
      _connectedEndpoints.add(endpoint);
      _mqttClient.updates!.listen((List<MqttReceivedMessage<MqttMessage?>>? messages) {
        final mqttMessage = messages![0].payload;
        if (mqttMessage! is MqttPublishMessage) return;
        final String payload =
            MqttPublishPayload.bytesToStringAsString((mqttMessage as MqttPublishMessage).payload.message);
        newI40MessageReceived(payload);
      });
    } catch (e) {
      notifyConsumingFailed(endpoint, S3IException('Failed to subscribe to $endpoint: $e'));
    }
  }

  /// Unsubscribe from a topic
  @override
  Future<void> stopConsuming(String endpoint) async {
    try {
      _mqttClient.unsubscribe(endpoint);
      _connectedEndpoints.remove(endpoint);
      print('Unsubscribed from $endpoint');
    } catch (e) {
      print('Failed to unsubscribe from $endpoint: $e');
    }
  }

  /// Reconnect to the broker
  @override
  Future<void> reconnectToBroker() async {
    _mqttClient.disconnect();
    await connectToBroker();
    await _reconnectEndpoints();
  }

  /// Check if the client is connected
  @override
  bool isConnected() {
    return _mqttClient.connectionStatus!.state == MqttConnectionState.connected;
  }

  /// Reconnect to previously connected endpoints
  Future<void> _reconnectEndpoints() async {
    for (final endpoint in _connectedEndpoints) {
      try {
        _mqttClient.subscribe(endpoint, MqttQos.atMostOnce);
        print('Reconnected to $endpoint');
      } catch (e) {
        print('Failed to reconnect to $endpoint: $e');
      }
    }
  }
}
