import 'dart:convert';

import 'package:crypto/crypto.dart';
import 'package:dart_amqp/dart_amqp.dart';
import 'package:flutter/foundation.dart';

//import 'package:s3i_flutter/src/broker/messages/event_system_messages.dart';
import 'package:s3i_flutter/src/broker/messages/i40_message.dart';

import 'package:s3i_flutter/s3i_flutter.dart';

// TODO(poq): if the broker rest api supports other exchanges -> update

/// The [EventSystemConnector] simplifies the usage of the S3I-Event-System.
///
/// Usable for only one event queue.
///
/// It's provides methods to connect/subscribe to the different events types
/// `namedEvents` and `customEvents` and parses the events messages.
class EventSystemConnector {
  /// Creates a new [EventSystemConnector] with a [s3iCore] and
  /// a [brokerInterface] connected to the "normal" exchange.
  ///
  /// The [brokerInterface] is optional because it's only used for custom
  /// events.
  EventSystemConnector(this.s3iCore, {this.brokerInterface}) {
    eventBrokerConnector = getActiveBrokerEventConnector(
      this.s3iCore.authManager,
      args: {
        'exchangeType': ExchangeType.TOPIC,
        'exchangeName': 'eventExchange',
        'brokerHost': '127.0.0.1', //TODO only for local testing
        // 'port': 5672, //port depends on the broker (amqp or mqtt)
        'virtualHost': 's3i',
      },
    );
    eventBrokerConnector
      ..subscribeI40EventMessageReceived((I40EventMessage event) {
        if (_eventCallbacks.containsKey(event.topic)) {
          for (final Function(I40EventMessage event) callbackF in _eventCallbacks[event.topic]!) {
            callbackF(event);
            removeUnhandled(event);
          }
        } else {
          addUnhandledEventMessage(event);
          _notifyEventWithoutCallbackReceived(event);
        }
      })
      ..subscribeConsumingFailed((String endpoint, Exception exception) {
        onErrorCallback(S3IException(exception.toString()));
      })
      ..setTopicBindingRecovery(() {
        subscribeNamedEvent(events: _eventCallbacks);
      });
  }

  /// The [S3ICore] used to connect to the Config REST API.
  final S3ICore s3iCore;

  /// The [ActiveBrokerInterface] used to communicate with the publisher
  /// via the normal direct exchange.
  final ActiveBrokerInterface? brokerInterface;

  /// This callback is invoked if an error occurs during subscription/message
  /// receiving.
  // ignore: prefer_function_declarations_over_variables
  final Function(S3IException exception) onErrorCallback = (S3IException exception) {};

  /// The broker interface for communication with the eventExchange.
  late final ActiveBrokerInterface eventBrokerConnector;

  /// Stores the topics of the subscribed events and the matching callbacks.
  Map<String, List<Function(I40EventMessage event)>> _eventCallbacks =
      <String, List<Function(I40EventMessage event)>>{};

  /// The endpoint to which the [eventBrokerConnector] is connected.
  Endpoint? _endpoint;

  /// All registered functions which are invoked if an [I40EventMessage] is
  /// received without a respective callback.
  final Set<Function(I40EventMessage)> _callbacksForEventWithoutCallback = <Function(I40EventMessage)>{};

  /// List of unhandled Messages
  List<I40EventMessage> unhandledEventMessages = [];

  /// creates an event queue with additional desired topic bindings
  Future<Endpoint> createEventQueue({Map<String, List<Function(I40EventMessage event)>> events = const {}}) async {
    final String thisThingId = s3iCore.authManager.clientIdentity.id;
    // create event queue with desired binding if not existing yet
    _eventCallbacks = events;
    _endpoint = await s3iCore.createEventQueueBinding(thisThingId, events.keys.toList());
    return _endpoint as Endpoint;
  }

  /// Starts receiving events
  Future<void> startConsumingEvents(String endpoint) async {
    _endpoint = Endpoint(endpoint);
    await eventBrokerConnector.startConsuming(endpoint);
  }

  /// Sets a queue binding to the given [eventTopic]
  Future<void> subscribeNamedEvent({required Map<String, List<Function(I40EventMessage event)>> events}) async {
    if(kIsWeb){
      _eventCallbacks = events;
      for(var eventTopic in events.keys){
        eventBrokerConnector.startConsuming(eventTopic);
      }
      return;
    }

    final String thisThingId = s3iCore.authManager.clientIdentity.id;
    // create event queue with desired binding if not existing yet
    var _endpoint = await s3iCore.createEventQueueBinding(thisThingId, events.keys.toList());
    print(_endpoint.endpoint);
    if (_endpoint != null) {
      _eventCallbacks = events;
      //await s3iCore.setEventQueueBinding(thisThingId, events.keys.toList().isEmpty? [] : events.keys.toList());

      await eventBrokerConnector.startConsuming(_endpoint.endpoint!);
    }
  }

  /// notify callbacks when event without callback is received
  void _notifyEventWithoutCallbackReceived(I40EventMessage message) {
    for (final Function(I40EventMessage) callback in _callbacksForEventWithoutCallback) {
      callback(message);
    }
  }

  /// Subscribes to all EventWithoutCallbackReceived-Events.
  void subscribeEventWithoutCallbackReceived(Function(I40EventMessage) callback) {
    _callbacksForEventWithoutCallback.add(callback);
  }

  void consumeUnhandledMessagesAgain() {
    var handledEventMessages = [];
    for (var eventMessage in unhandledEventMessages) {
      if (_eventCallbacks.containsKey(eventMessage.topic)) {
        for (final Function(I40EventMessage event) callbackF in _eventCallbacks[eventMessage.topic]!) {
          callbackF(eventMessage);
          handledEventMessages.add(eventMessage);
        }
      } else {
        addUnhandledEventMessage(eventMessage);
        _notifyEventWithoutCallbackReceived(eventMessage);
      }
    }
    for (var message in handledEventMessages) {
      removeUnhandled(message);
    }
  }

  void addUnhandledEventMessage(I40EventMessage eventMessage) {
    if (unhandledEventMessages.contains(eventMessage)) {
      return;
    } else {
      unhandledEventMessages.add(eventMessage);
    }
  }

  void removeUnhandled(I40EventMessage eventMessage) {
    unhandledEventMessages.remove(eventMessage);
  }

  // TODO(bek): add method for unsubscribe specific events

  /// Shuts down the connection to the broker. Currently the broker queue is
  /// deleted too (the REST-API creates queues with `auto_delete = true`).
  void stopConsumingEvents() {
    if (_endpoint != null) {
      eventBrokerConnector.stopConsuming(_endpoint!.endpoint);
      _eventCallbacks.clear();
    }
  }

  /// Publish event message via EventExchange
  void publishEventMessage({required I40EventMessage eventMessage}) {
    eventBrokerConnector.sendI40Message(eventMessage, {Endpoint(eventMessage.topic)});
  }
}
