import 'package:collection/collection.dart';
import 'package:dart_aas/dart_aas.dart';
import 'package:s3i_flutter/src/exceptions/invalid_json_schema_exception.dart';
import 'package:s3i_flutter/src/exceptions/json_missing_key_exception.dart';
import 'package:s3i_flutter/src/utils/json_keys.dart';

enum I40MessageType { unkown, request, reply, event }

enum ConversationRole { unkown, requester, replier, emitter }

/// Defines a abstract class for a ConversationPartner of a [Frame] within a [I40Message].
abstract class ConversationPartner {
  /// Creates a [ConversationPartner] based on the [identification] and [conversationRole].
  ConversationPartner(
      {this.identification = '',
      this.conversationRole = ConversationRole.unkown});

  /// Creates Subclasses form the [ConversationPartner] class  with the information stored in the [json]
  ///
  /// Throws a [JsonMissingKeyException] if there is missing one of the needed
  /// keys ([I40MessageKeys.identification],[I40MessageKeys.role],[I40MessageKeys.name]) in the [json].
  /// Throws a [InvalidJsonSchemaException] if some values
  /// doesn't match the expected value type.
  void generatesFromJson(Map<String, dynamic> json) {
    try {
      identification = json.containsKey(I40MessageKeys.identification)
          ? json[I40MessageKeys.identification] as String
          : throw JsonMissingKeyException(
              I40MessageKeys.identification, json.toString());

      final Map<String, dynamic> roleDict = json
              .containsKey(I40MessageKeys.role)
          ? json[I40MessageKeys.role] as Map<String, dynamic>
          : throw JsonMissingKeyException(I40MessageKeys.role, json.toString());

      final String roleName = roleDict.containsKey(I40MessageKeys.name)
          ? roleDict[I40MessageKeys.name] as String
          : throw JsonMissingKeyException(
              I40MessageKeys.name, roleDict.toString());

      if (roleName == ConversationRole.requester.toString()) {
        conversationRole = ConversationRole.requester;
      } else if (roleName == ConversationRole.replier.toString()) {
        conversationRole = ConversationRole.replier;
      } else {
        conversationRole = ConversationRole.emitter;
      }
    } on TypeError catch (e) {
      throw InvalidJsonSchemaException(
          e.stackTrace.toString(), json.toString());
    }
  }

  /// S3I-Identitiy of the Conversationpartner
  String identification;

  /// Conversationrole of the Conversationpartner depending on the message type
  ConversationRole conversationRole;

  /// Deserializes [ConversationPartner] class  into a [Map]
  Map<String, dynamic> toJson() {
    return <String, dynamic>{
      I40MessageKeys.identification: identification,
      I40MessageKeys.role: <String, String>{
        I40MessageKeys.name: conversationRole.name.toString()
      }
    };
  }
}

/// Defines the [Sender] class within a [Frame].
class Sender extends ConversationPartner {
  /// Creates a [Sender] based on the [identification] and [conversationRole].
  Sender(
      {String identification = '',
      ConversationRole conversationRole = ConversationRole.unkown})
      : super(
            identification: identification, conversationRole: conversationRole);

  /// Creates a [Sender] class from the information stored in the [json]
  ///
  /// Throws a [JsonMissingKeyException] if there is missing one of the needed
  /// keys ([I40MessageKeys.identification],[I40MessageKeys.role],[I40MessageKeys.name]) in the [json].
  /// Throws a [InvalidJsonSchemaException] if some values
  /// doesn't match the expected value type.
  factory Sender.fromJson(Map<String, dynamic> json) {
    return Sender()..generatesFromJson(json);
  }
}

/// Defines the [Receiver] class within a [Frame].
class Receiver extends ConversationPartner {
  /// Creates a [Receiver] based on the [identification] and [conversationRole].
  Receiver(
      {String identification = '',
      ConversationRole conversationRole = ConversationRole.unkown})
      : super(
            identification: identification, conversationRole: conversationRole);

  /// Creates a [Receiver] class from the information stored in the [json]
  ///
  /// Throws a [JsonMissingKeyException] if there is missing one of the needed
  /// keys ([I40MessageKeys.identification],[I40MessageKeys.role],[I40MessageKeys.name]) in the [json].
  /// Throws a [InvalidJsonSchemaException] if some values
  /// doesn't match the expected value type.
  factory Receiver.fromJson(Map<String, dynamic> json) {
    return Receiver()..generatesFromJson(json);
  }
}

/// Defines the [Frame] within a [I40Message].
class Frame {
  Reference semanticProtocol;
  I40MessageType messageType;
  String messageId;
  Sender sender;
  Receiver? receiver;
  String? inReplyTo;
  String? replyBy;
  String? conversationId;

  /// Creates a [Frame].
  Frame({
    required this.semanticProtocol,
    required this.messageType,
    required this.messageId,
    required this.sender,
    this.receiver,
    this.inReplyTo,
    this.replyBy,
    this.conversationId,
  });

  /// Creates a [Frame] class from the information stored in the [json]
  ///
  /// Throws a [JsonMissingKeyException] if there is missing one of the needed
  /// keys ([I40MessageKeys.semanticId],[I40MessageKeys.type],
  /// [I40MessageKeys.messageId],[I40MessageKeys.sender],[I40MessageKeys.receiver])
  /// in the [json].
  /// Throws a [InvalidJsonSchemaException] if some values
  /// doesn't match the expected value type.
  factory Frame.fromJson(Map<String, dynamic> json) {
    try {
      final Reference semanticProtocol =
          json.containsKey(I40MessageKeys.semanticProtocol)
              ? Reference.fromJson(
                  json[I40MessageKeys.semanticProtocol] as Map<String, dynamic>)
              : throw JsonMissingKeyException(
                  I40MessageKeys.semanticProtocol, json.toString());

      final I40MessageType messageType = json.containsKey(I40MessageKeys.type)
          ? I40MessageType.values.firstWhereOrNull(
                  (I40MessageType messageType) =>
                      messageType.name == json[I40MessageKeys.type]) ??
              I40MessageType.unkown
          : throw JsonMissingKeyException(AASKeys.semanticId, json.toString());

      final String messageId = json.containsKey(I40MessageKeys.messageId)
          ? json[I40MessageKeys.messageId]
          : throw JsonMissingKeyException(
              I40MessageKeys.messageId, json.toString());

      final Sender sender = json.containsKey(I40MessageKeys.sender)
          ? Sender.fromJson(json[I40MessageKeys.sender])
          : throw JsonMissingKeyException(
              I40MessageKeys.sender, json.toString());

      final Receiver? receiver = json.containsKey(I40MessageKeys.receiver)
          ? Receiver.fromJson(json[I40MessageKeys.receiver])
          : null;

      final String? inReplyTo = json.containsKey(I40MessageKeys.inReplyTo)
          ? json[I40MessageKeys.inReplyTo] as String
          : null;

      final String? replyBy = json.containsKey(I40MessageKeys.replyBy)
          ? json[I40MessageKeys.replyBy] as String
          : null;

      final String? conversationId =
          json.containsKey(I40MessageKeys.conversationId)
              ? json[I40MessageKeys.conversationId] as String
              : null;

      return Frame(
        semanticProtocol: semanticProtocol,
        messageType: messageType,
        messageId: messageId,
        sender: sender,
        receiver: receiver,
        inReplyTo: inReplyTo,
        replyBy: replyBy,
        conversationId: conversationId,
      );
    } on TypeError catch (e) {
      throw InvalidJsonSchemaException(
          e.stackTrace.toString(), json.toString());
    }
  }

  /// Deserializes a Frame into a Json represenation as a [Map].
  Map<String, dynamic> toJson() {
    final Map<String, dynamic> json = {};
    json[I40MessageKeys.semanticProtocol] = semanticProtocol.toJson();
    json[I40MessageKeys.type] = messageType.name;
    json[I40MessageKeys.messageId] = messageId;
    json[I40MessageKeys.sender] = sender.toJson();
    if(receiver != null) json[I40MessageKeys.receiver] = receiver!.toJson();
    if (inReplyTo != null) json[I40MessageKeys.inReplyTo] = inReplyTo;
    if (replyBy != null) json[I40MessageKeys.replyBy] = replyBy;
    if (conversationId != null) {
      json[I40MessageKeys.conversationId] = conversationId;
    }
    return json;
  }
}
