import 'dart:core';
import 'dart:convert';

import 'package:dart_aas/dart_aas.dart';
import 'package:s3i_flutter/src/broker/messages/frame.dart';
import 'package:s3i_flutter/src/exceptions/invalid_json_schema_exception.dart';
import 'package:s3i_flutter/src/exceptions/invalid_message_json_value.dart';
import 'package:s3i_flutter/src/exceptions/json_missing_key_exception.dart';
import 'package:s3i_flutter/src/utils/json_keys.dart';
import 'package:uuid/uuid.dart';

class I40Message {
  Frame? frame;
  List<SubmodelElement> interactionElements;

  I40Message({this.frame, this.interactionElements = const <SubmodelElement>[]});

  static I40Message withFrameElements(Frame frame, List<SubmodelElement> elements) {
    var i40Message = I40Message()
      ..frame = frame
      ..interactionElements = elements;
    return i40Message;
  }

  factory I40Message.fromJson(Map<String, dynamic> json) {
    try {
      final Frame frame = json.containsKey(I40MessageKeys.frame)
          ? Frame.fromJson(json[I40MessageKeys.frame])
          : throw JsonMissingKeyException(I40MessageKeys.frame, json.toString());

      if (frame.messageType == I40MessageType.request) {
        if (frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.invokeOperationSync) ||
            frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.invokeOperationAsync)) {
          return I40InvokeOperationRequest.fromJson(json);
        } else if (frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.getSubmodelElementByPath)) {
          return I40GetSubmodelElementByPathRequest.fromJson(json);
        } else {
          throw InvalidMesssageJsonValue(
              'Message Type is request but no suitable semantic id found', frame.semanticProtocol.toString());
        }
      } else if (frame.messageType == I40MessageType.reply) {
        if (frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.invokeOperationSync) ||
            frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.invokeOperationAsync)) {
          return I40InvokeOperationReply.fromJson(json);
        } else if (frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.getSubmodelElementByPath)) {
          return I40GetSubmodelElementByPathReply.fromJson(json);
        } else {
          throw InvalidMesssageJsonValue(
              'Message Type is reply but no suitable semantic id found', frame.semanticProtocol.toString());
        }
      } else if (frame.messageType == I40MessageType.event) {
        return I40EventMessage.fromJson(json);
      } else {
        //Unknown
        throw InvalidMesssageJsonValue('Message Type unkonwn', frame.messageType.toString());
      }
    } on TypeError catch (e) {
      throw InvalidJsonSchemaException(e.stackTrace.toString(), json.toString());
    }
  }

  Map<String, dynamic> toJson() {
    final Map<String, dynamic> msgJson = <String, dynamic>{};
    if (frame != null) {
      msgJson[I40MessageKeys.frame] = frame!.toJson();
    }
    if (interactionElements.isNotEmpty) {
      final List<Map<String, dynamic>> elements = <Map<String, dynamic>>[];
      for (final SubmodelElement element in interactionElements) {
        elements.add(element.toJson());
      }
      msgJson[I40MessageKeys.interactionElements] = elements;
    }
    return msgJson;
  }
}

class I40InvokeOperationRequest extends I40Message {
  String submodelId;
  String operationIdShortPath;
  String sender;
  String receiver;
  String? messageId;
  String? inReplyTo;
  String? replyBy;
  String? conversationId;
  List<SubmodelElement> inputVariables;
  bool sync;

  I40InvokeOperationRequest(
      {required this.submodelId,
      required this.operationIdShortPath,
      required this.sender,
      required this.receiver,
      this.messageId,
      this.inReplyTo,
      this.replyBy,
      this.conversationId,
      required this.inputVariables,
      required this.sync}) {
    final ReferenceKey reference =
        sync ? I40MessageGlobalReferenceKeys.invokeOperationSync : I40MessageGlobalReferenceKeys.invokeOperationAsync;
    final Reference semanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [reference],
    );
    const I40MessageType messageType = I40MessageType.request;
    messageId ??= const Uuid().v4();
    final Sender senderObj = Sender(identification: sender, conversationRole: ConversationRole.requester);
    final Receiver receiverObj = Receiver(identification: receiver, conversationRole: ConversationRole.replier);
    final Frame frame = Frame(
      semanticProtocol: semanticProtocol,
      messageType: messageType,
      messageId: messageId!,
      sender: senderObj,
      receiver: receiverObj,
      inReplyTo: inReplyTo,
      replyBy: replyBy,
      conversationId: conversationId,
    );

    final List<SubmodelElement> elements = <SubmodelElement>[];
    final ReferenceKey submodelIdReference = sync
        ? I40MessageGlobalReferenceKeys.invokeOperationSyncSubmodelId
        : I40MessageGlobalReferenceKeys.invokeOperationAsyncSubmodelId;
    final Property<String> submodelIdElement = Property<String>(
      idShort: 'submodelId',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: [submodelIdReference],
      ),
      value: submodelId,
    );
    elements.add(submodelIdElement);

    final ReferenceKey pathReference = sync
        ? I40MessageGlobalReferenceKeys.invokeOperationSyncPath
        : I40MessageGlobalReferenceKeys.invokeOperationAsyncPath;
    final Property<String> pathElement = Property<String>(
      idShort: 'path',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: <ReferenceKey>[pathReference],
      ),
      value: operationIdShortPath,
    );
    elements.add(pathElement);

    final ReferenceKey inputArgumentReference = sync
        ? I40MessageGlobalReferenceKeys.invokeOperationSyncInputArgument
        : I40MessageGlobalReferenceKeys.invokeOperationAsyncInputArgument;
    final SubmodelElementCollection inputArgumentElement = SubmodelElementCollection(
      idShort: 'inputArgument',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: <ReferenceKey>[inputArgumentReference],
      ),
    )..elements = inputVariables;

    elements.add(inputArgumentElement);
    super.frame = frame;
    super.interactionElements = elements;
  }

  factory I40InvokeOperationRequest.fromJson(Map<String, dynamic> json) {
    String? submodelId;
    String? operationIdShortPath;
    Frame frame;
    List<SubmodelElement>? inputVariables;
    bool? sync;

    try {
      frame = json.containsKey(I40MessageKeys.frame)
          ? Frame.fromJson(json[I40MessageKeys.frame])
          : throw JsonMissingKeyException(I40MessageKeys.frame, json.toString());

      if (frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.invokeOperationSync)) {
        sync = true;
      } else if (frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.invokeOperationAsync)) {
        sync = false;
      } else {
        throw InvalidMesssageJsonValue('Invalid semantic reference for invoke operation request', json.toString());
      }

      final List<SubmodelElement> interactionElements = json.containsKey(I40MessageKeys.interactionElements)
          ? createSubmodelElementListFromJson(json[I40MessageKeys.interactionElements] as List<dynamic>)
          : throw JsonMissingKeyException(I40MessageKeys.interactionElements, json.toString());

      for (final SubmodelElement element in interactionElements) {
        if (element is Property && element.idShort == 'submodelId') {
          submodelId = element.value as String;
        } else if (element is Property && element.idShort == 'path') {
          operationIdShortPath = element.value as String;
        } else if (element is SubmodelElementCollection && element.idShort == 'inputArgument') {
          inputVariables = element.elements;
        }
      }

      if (submodelId == null || operationIdShortPath == null || inputVariables == null) {
        throw InvalidMesssageJsonValue(
            'No InteractionElement with one of the following shortIds provided [submodelId, operationIdShortPath, inputVariables]',
            jsonEncode(json[I40MessageKeys.interactionElements]));
      }

      if(frame.receiver == null) {
        throw JsonMissingKeyException(I40MessageKeys.receiver, json.toString());
      }

      if (sync) {
        return I40InvokeOperationSyncRequest(
          submodelId: submodelId,
          operationIdShortPath: operationIdShortPath,
          sender: frame.sender.identification,
          receiver: frame.receiver!.identification,
          messageId: frame.messageId,
          inReplyTo: frame.inReplyTo,
          replyBy: frame.replyBy,
          conversationId: frame.conversationId,
          inputVariables: inputVariables,
        );
      } else {
        return I40InvokeOperationAsyncRequest(
          submodelId: submodelId,
          operationIdShortPath: operationIdShortPath,
          sender: frame.sender.identification,
          receiver: frame.receiver!.identification,
          messageId: frame.messageId,
          inReplyTo: frame.inReplyTo,
          replyBy: frame.replyBy,
          conversationId: frame.conversationId,
          inputVariables: inputVariables,
        );
      }
    } on TypeError catch (e) {
      throw InvalidJsonSchemaException(e.stackTrace.toString(), json.toString());
    }
  }
}

class I40InvokeOperationSyncRequest extends I40InvokeOperationRequest {
  I40InvokeOperationSyncRequest({
    required super.submodelId,
    required super.operationIdShortPath,
    required super.sender,
    required super.receiver,
    super.messageId,
    super.inReplyTo,
    super.replyBy,
    super.conversationId,
    required super.inputVariables,
  }) : super(sync: true);
}

class I40InvokeOperationAsyncRequest extends I40InvokeOperationRequest {
  I40InvokeOperationAsyncRequest({
    required super.submodelId,
    required super.operationIdShortPath,
    required super.sender,
    required super.receiver,
    super.messageId,
    super.inReplyTo,
    super.replyBy,
    super.conversationId,
    required super.inputVariables,
  }) : super(sync: false);
}

class I40InvokeOperationReply extends I40Message {
  String sender;
  String receiver;
  String? messageId;
  String? inReplyTo;
  String? replyBy;
  String? conversationId;
  List<SubmodelElement> outputVariables;
  bool sync;
  bool success;
  String statusCode;
  String errorMessage;

  I40InvokeOperationReply(
      {required this.sender,
      required this.receiver,
      this.messageId,
      this.inReplyTo,
      this.replyBy,
      this.conversationId,
      required this.outputVariables,
      required this.sync,
      required this.success,
      this.statusCode = '',
      this.errorMessage = ''}) {
    if (success) {
      statusCode = 'Success';
      errorMessage = '';
    }

    final ReferenceKey reference =
        sync ? I40MessageGlobalReferenceKeys.invokeOperationSync : I40MessageGlobalReferenceKeys.invokeOperationAsync;
    final Reference semanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: <ReferenceKey>[reference],
    );

    const I40MessageType messageType = I40MessageType.reply;
    messageId ??= const Uuid().v4();
    final Sender senderObj = Sender(identification: sender, conversationRole: ConversationRole.replier);
    final Receiver receiverObj = Receiver(identification: receiver, conversationRole: ConversationRole.requester);

    final Frame frame = Frame(
      semanticProtocol: semanticProtocol,
      messageType: messageType,
      messageId: messageId!,
      sender: senderObj,
      receiver: receiverObj,
      inReplyTo: inReplyTo,
      replyBy: replyBy,
      conversationId: conversationId,
    );

    final List<SubmodelElement> elements = <SubmodelElement>[];

    final ReferenceKey statusCodeReference = sync
        ? I40MessageGlobalReferenceKeys.invokeOperationSyncStatusCode
        : I40MessageGlobalReferenceKeys.invokeOperationAsyncStatusCode;
    final Property<String> statusCodeElement = Property<String>(
      idShort: 'statusCode',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: <ReferenceKey>[statusCodeReference],
      ),
      value: statusCode,
    );
    elements.add(statusCodeElement);

    /// Add payload interaction element based on success of operation
    late SubmodelElementCollection payloadElement;
    if (success) {
      payloadElement = __createSuccessPayload(outputVariables, sync);
    } else {
      payloadElement = __createErrorPayload(errorMessage, sync);
    }

    elements.add(payloadElement);

    super.frame = frame;
    super.interactionElements = elements;
  }

  SubmodelElementCollection __createSuccessPayload(List<SubmodelElement> outputVariables, bool sync) {
    final ReferenceKey payloadReferenceKey = sync
        ? I40MessageGlobalReferenceKeys.invokeOperationSyncPayload
        : I40MessageGlobalReferenceKeys.invokeOperationAsyncPayload;

    final SubmodelElementCollection payloadElement = SubmodelElementCollection(
        idShort: 'payload',
        semanticId: Reference(
          type: ReferenceTypes.ExternalReference,
          keys: [payloadReferenceKey],
        ));

    /// OutputArgument Element
    final ReferenceKey outputArgumentReference = sync
        ? I40MessageGlobalReferenceKeys.invokeOperationSyncOutputArgument
        : I40MessageGlobalReferenceKeys.invokeOperationAsyncOutputArgument;
    final SubmodelElementCollection outputArgumentElement = SubmodelElementCollection(
      idShort: 'outputArguments',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: [outputArgumentReference],
      ),
    );
    outputArgumentElement.elements = outputVariables;

    payloadElement.addElement(outputArgumentElement);

    /// ExecutionState Element
    final ReferenceKey executionStateReferenceKey = sync
        ? I40MessageGlobalReferenceKeys.invokeOperationSyncExecutionState
        : I40MessageGlobalReferenceKeys.invokeOperationAsyncExecutionState;
    final Property<String> executionStateElement = Property<String>(
      idShort: 'executionState',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: [executionStateReferenceKey],
      ),
      value: 'Completed',
    );
    payloadElement.addElement(executionStateElement);

    /// Success Element
    final ReferenceKey successReferenceKey = sync
        ? I40MessageGlobalReferenceKeys.invokeOperationSyncSuccess
        : I40MessageGlobalReferenceKeys.invokeOperationAsyncSuccess;
    final Property<bool> successElement = Property<bool>(
      idShort: 'success',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: [successReferenceKey],
      ),
      value: true,
    );
    payloadElement.addElement(successElement);

    return payloadElement;
  }

  SubmodelElementCollection __createErrorPayload(String errorMessage, bool sync) {
    final ReferenceKey payloadReferenceKey = sync
        ? I40MessageGlobalReferenceKeys.invokeOperationSyncPayload
        : I40MessageGlobalReferenceKeys.invokeOperationAsyncPayload;

    final SubmodelElementCollection payloadElement = SubmodelElementCollection(
        idShort: 'payload',
        semanticId: Reference(
          type: ReferenceTypes.ExternalReference,
          keys: [payloadReferenceKey],
        ));

    /// Result Element
    final SubmodelElementCollection resultElement = SubmodelElementCollection(
      idShort: 'result',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: [I40MessageGlobalReferenceKeys.result],
      ),
    );

    /// Result Element
    final SubmodelElementCollection messageElement = SubmodelElementCollection(
      idShort: 'message',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: [I40MessageGlobalReferenceKeys.message],
      ),
    );

    /// MessageType Element
    final Property<String> messageTypeElement = Property<String>(
      idShort: 'success',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: [I40MessageGlobalReferenceKeys.messageType],
      ),
      value: 'Error',
    );
    messageElement.addElement(messageTypeElement);

    /// Text Element
    final Property<String> textElement = Property<String>(
      idShort: 'text',
      semanticId: Reference(
        type: ReferenceTypes.ExternalReference,
        keys: [I40MessageGlobalReferenceKeys.text],
      ),
      value: errorMessage,
    );
    messageElement.addElement(textElement);
    resultElement.addElement(messageElement);
    payloadElement.addElement(resultElement);

    return payloadElement;
  }

  factory I40InvokeOperationReply.fromJson(Map<String, dynamic> json) {
    Frame frame;
    List<SubmodelElement>? outputVariables;
    bool? sync;
    bool? success;
    String? statusCode;
    String? errorMessage;

    try {
      frame = json.containsKey(I40MessageKeys.frame)
          ? Frame.fromJson(json[I40MessageKeys.frame])
          : throw JsonMissingKeyException(I40MessageKeys.frame, json.toString());

      if (frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.invokeOperationSync)) {
        sync = true;
      } else if (frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.invokeOperationAsync)) {
        sync = false;
      } else {
        throw InvalidMesssageJsonValue('Invalid semantic reference for invoke operation request', json.toString());
      }

      final List<SubmodelElement> interactionElements = json.containsKey(I40MessageKeys.interactionElements)
          ? createSubmodelElementListFromJson(json[I40MessageKeys.interactionElements] as List<dynamic>)
          : throw JsonMissingKeyException(I40MessageKeys.interactionElements, json.toString());

      late SubmodelElementCollection _payloadElement;
      for (final SubmodelElement element in interactionElements) {
        if (element is SubmodelElementCollection && element.idShort == 'payload') {
          _payloadElement = element;
        } else if (element is Property && element.idShort == 'statusCode') {
          statusCode = element.value as String;
        }
      }

      if (statusCode == 'Success') {
        success = true;
        errorMessage = '';
        outputVariables = (_payloadElement.elements
                    .where((e) => e is SubmodelElementCollection && e.idShort == 'outputArguments')
                    .firstOrNull as SubmodelElementCollection?)
                ?.elements ??
            [];
      } else {
        success = false;
        outputVariables = <SubmodelElement>[];
        errorMessage = '';
        final SubmodelElementCollection result = _payloadElement.elements.first as SubmodelElementCollection;
        final SubmodelElementCollection message = result.elements.first as SubmodelElementCollection;
        for (final SubmodelElement el in message.elements) {
          if (el is Property && el.idShort == 'text') {
            errorMessage = el.value as String;
          }
        }
      }

      if (outputVariables == null || statusCode == null || errorMessage == null) {
        throw InvalidMesssageJsonValue(
            'No InteractionElement with one of the following shortIds provided [outputVariables, statusCode, errorMessage]',
            json[I40MessageKeys.interactionElements].toString());
      }

      if(frame.receiver == null) {
        throw JsonMissingKeyException(I40MessageKeys.receiver, json.toString());
      }

      if (sync) {
        return I40InvokeOperationSyncReply(
          sender: frame.sender.identification,
          receiver: frame.receiver!.identification,
          messageId: frame.messageId,
          inReplyTo: frame.inReplyTo,
          replyBy: frame.replyBy,
          conversationId: frame.conversationId,
          outputVariables: outputVariables,
          success: success,
          statusCode: statusCode,
          errorMessage: errorMessage,
        );
      } else {
        return I40InvokeOperationAsyncReply(
          sender: frame.sender.identification,
          receiver: frame.receiver!.identification,
          messageId: frame.messageId,
          inReplyTo: frame.inReplyTo,
          replyBy: frame.replyBy,
          conversationId: frame.conversationId,
          outputVariables: outputVariables,
          success: success,
          statusCode: statusCode,
          errorMessage: errorMessage,
        );
      }
    } on TypeError catch (e) {
      throw InvalidJsonSchemaException(e.stackTrace.toString(), json.toString());
    }
  }
}

class I40InvokeOperationSyncReply extends I40InvokeOperationReply {
  I40InvokeOperationSyncReply(
      {required String sender,
      required String receiver,
      required bool success,
      super.messageId,
      super.inReplyTo,
      super.replyBy,
      super.conversationId,
      List<SubmodelElement>? outputVariables,
      String statusCode = '',
      String errorMessage = ''})
      : super(
            sender: sender,
            receiver: receiver,
            outputVariables: outputVariables ?? <SubmodelElement>[],
            sync: true,
            success: success,
            statusCode: statusCode,
            errorMessage: errorMessage);
}

class I40InvokeOperationAsyncReply extends I40InvokeOperationReply {
  I40InvokeOperationAsyncReply(
      {required String sender,
      required String receiver,
      super.messageId,
      super.inReplyTo,
      super.replyBy,
      super.conversationId,
      required bool success,
      List<SubmodelElement>? outputVariables,
      String statusCode = '',
      String errorMessage = ''})
      : super(
            sender: sender,
            receiver: receiver,
            outputVariables: outputVariables ?? <SubmodelElement>[],
            sync: false,
            success: success,
            statusCode: statusCode,
            errorMessage: errorMessage);
}

class I40GetSubmodelElementByPathRequest extends I40Message {
  String submodelId;
  String submodelElementIdShortPath;
  String sender;
  String receiver;
  String? messageId;
  String? inReplyTo;
  String? replyBy;
  String? conversationId;

  I40GetSubmodelElementByPathRequest({
    required this.submodelId,
    required this.submodelElementIdShortPath,
    required this.sender,
    required this.receiver,
    this.messageId,
    this.inReplyTo,
    this.replyBy,
    this.conversationId,
  }) {
    final Reference frameSemanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [I40MessageGlobalReferenceKeys.getSubmodelElementByPath],
    );

    const I40MessageType messageType = I40MessageType.request;
    messageId ??= Uuid().v4();
    final Sender senderObj = Sender(
      identification: sender,
      conversationRole: ConversationRole.requester,
    );
    final Receiver receiverObj = Receiver(
      identification: receiver,
      conversationRole: ConversationRole.replier,
    );

    final Frame frame = Frame(
      semanticProtocol: frameSemanticProtocol,
      messageType: messageType,
      messageId: messageId!,
      sender: senderObj,
      receiver: receiverObj,
      inReplyTo: inReplyTo,
      replyBy: replyBy,
      conversationId: conversationId,
    );

    final List<Property<String>> elements = <Property<String>>[];

    // status code
    final Reference submodelIdSemanticId = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [
        I40MessageGlobalReferenceKeys.getSubmodelElementByPathSubmodelId,
      ],
    );
    final Property<String> submodelIdElement = Property<String>(
      idShort: 'submodelId',
      semanticId: submodelIdSemanticId,
      value: submodelId,
    );
    elements.add(submodelIdElement);

    // path
    final Reference pathSemanticId = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [
        I40MessageGlobalReferenceKeys.getSubmodelElementByPathPath,
      ],
    );
    final Property<String> pathElement = Property<String>(
      idShort: 'path',
      semanticId: pathSemanticId,
      value: submodelElementIdShortPath,
    );
    elements.add(pathElement);

    super.frame = frame;
    super.interactionElements = elements;
  }

  factory I40GetSubmodelElementByPathRequest.fromJson(Map<String, dynamic> json) {
    String? sender;
    String? receiver;
    String? submodelId;
    String? submodelElementIdShortPath;

    try {
      final Frame frame = json.containsKey(I40MessageKeys.frame)
          ? Frame.fromJson(json[I40MessageKeys.frame])
          : throw JsonMissingKeyException(I40MessageKeys.frame, json.toString());

      if (frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.getSubmodelElementByPath) == false) {
        throw InvalidMesssageJsonValue(
            'Invalid semantic reference for GetSubmodelElementByPath request', json.toString());
      }

      sender = frame.sender.identification;
      receiver = frame.receiver?.identification;
      if(receiver == null) {
        throw JsonMissingKeyException(I40MessageKeys.receiver, json.toString());
      }

      final List<SubmodelElement> interactionElements = json.containsKey(I40MessageKeys.interactionElements)
          ? createSubmodelElementListFromJson(json[I40MessageKeys.interactionElements] as List<dynamic>)
          : throw JsonMissingKeyException(I40MessageKeys.interactionElements, json.toString());

      late SubmodelElementCollection _submodeleElementElement;
      for (final SubmodelElement element in interactionElements) {
        if (element is SubmodelElementCollection && element.idShort == 'submodelElement') {
          _submodeleElementElement = element;
        } else if (element is Property && element.idShort == 'submodelId') {
          submodelId = element.value as String;
        } else if (element is Property && element.idShort == 'path') {
          submodelElementIdShortPath = element.value as String;
        }
      }

      if (submodelId == null || submodelElementIdShortPath == null) {
        throw InvalidMesssageJsonValue(
            'No InteractionElement with one of the following shortIds provided [submodelId, path]',
            json[I40MessageKeys.interactionElements].toString());
      }

      return I40GetSubmodelElementByPathRequest(
        submodelId: submodelId,
        submodelElementIdShortPath: submodelElementIdShortPath,
        sender: sender,
        receiver: receiver,
        messageId: frame.messageId,
        inReplyTo: frame.inReplyTo,
        replyBy: frame.replyBy,
        conversationId: frame.conversationId,
      );
    } on TypeError catch (e) {
      throw InvalidJsonSchemaException(e.stackTrace.toString(), json.toString());
    }
  }
}

class I40GetSubmodelElementByPathReply extends I40Message {
  SubmodelElement submodelElement;
  String sender;
  String receiver;
  String? messageId;
  String? inReplyTo;
  String? replyBy;
  String? conversationId;

  I40GetSubmodelElementByPathReply({
    required this.submodelElement,
    required this.sender,
    required this.receiver,
    this.messageId,
    this.inReplyTo,
    this.replyBy,
    this.conversationId,
  }) {
    final Reference frameSemanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [I40MessageGlobalReferenceKeys.getSubmodelElementByPath],
    );

    const I40MessageType messageType = I40MessageType.reply;
    messageId ??= const Uuid().v4();
    final Sender senderObj = Sender(
      identification: sender,
      conversationRole: ConversationRole.replier,
    );
    final Receiver receiverObj = Receiver(
      identification: receiver,
      conversationRole: ConversationRole.requester,
    );

    final Frame frame = Frame(
      semanticProtocol: frameSemanticProtocol,
      messageType: messageType,
      messageId: messageId!,
      sender: senderObj,
      receiver: receiverObj,
      inReplyTo: inReplyTo,
      replyBy: replyBy,
      conversationId: conversationId,
    );

    final List<SubmodelElement> elements = <SubmodelElement>[];

    // status code
    final Reference statusCodeSemanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [
        I40MessageGlobalReferenceKeys.getSubmodelElementByPathStatusCode,
      ],
    );
    final Property<String> statusCodeElement = Property<String>(
      idShort: 'statusCode',
      semanticId: statusCodeSemanticProtocol,
      value: 'Success',
    );
    elements.add(statusCodeElement);

    // payload
    final Reference payloadSemanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [
        I40MessageGlobalReferenceKeys.getSubmodelElementByPathStatusCode,
      ],
    );
    final SubmodelElementCollection payloadElement = SubmodelElementCollection(
      idShort: 'payload',
      semanticId: payloadSemanticProtocol,
    )..addElement(submodelElement); // TODOadd error case

    elements.add(payloadElement);

    super.frame = frame;
    super.interactionElements = elements;
  }

  factory I40GetSubmodelElementByPathReply.fromJson(Map<String, dynamic> json) {
    String? sender;
    String? receiver;
    SubmodelElement? submodelElement;
    String? statusCode;

    try {
      final Frame frame = json.containsKey(I40MessageKeys.frame)
          ? Frame.fromJson(json[I40MessageKeys.frame])
          : throw JsonMissingKeyException(I40MessageKeys.frame, json.toString());

      if (frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.getSubmodelElementByPath) == false) {
        throw InvalidMesssageJsonValue(
            'Invalid semantic reference for GetSubmodelElementByPath request', json.toString());
      }

      sender = frame.sender.identification;
      receiver = frame.receiver?.identification;
      if(receiver == null) {
        throw JsonMissingKeyException(I40MessageKeys.receiver, json.toString());
      }

      final List<SubmodelElement> interactionElements = json.containsKey(I40MessageKeys.interactionElements)
          ? createSubmodelElementListFromJson(json[I40MessageKeys.interactionElements] as List<dynamic>)
          : throw JsonMissingKeyException(I40MessageKeys.interactionElements, json.toString());

      SubmodelElementCollection? payloadElement;
      for (final SubmodelElement element in interactionElements) {
        if (element is SubmodelElementCollection && element.idShort == 'payload') {
          payloadElement = element;
        } else if (element is Property && element.idShort == 'statusCode') {
          statusCode = element.value as String;
        }
      }

      if (payloadElement == null || statusCode == null) {
        throw InvalidMesssageJsonValue(
            'No InteractionElement with one of the following shortIds provided [payload, statusCode]',
            json[I40MessageKeys.interactionElements].toString());
      }

      // get actual submodel element
      submodelElement = payloadElement.elements.firstOrNull;
      if (submodelElement == null) {
        throw InvalidMesssageJsonValue(
            'No SubmodelElement as part of the payload provided', payloadElement.toJson().toString());
      }

      return I40GetSubmodelElementByPathReply(
        submodelElement: submodelElement,
        sender: sender,
        receiver: receiver,
        messageId: frame.messageId,
        inReplyTo: frame.inReplyTo,
        replyBy: frame.replyBy,
        conversationId: frame.conversationId,
      );
    } on TypeError catch (e) {
      throw InvalidJsonSchemaException(e.stackTrace.toString(), json.toString());
    }
  }
}

class I40EventMessage extends I40Message {
  String sender;
  String? messageId;
  Reference eventModelReference;
  Reference? eventSemanticId;
  Reference observableModelReference;
  Reference? observableSemanticId;
  String topic;
  DateTime timestamp;
  SubmodelElement payload;

  I40EventMessage({
    required this.sender,
    this.messageId,
    required this.eventModelReference,
    this.eventSemanticId,
    required this.observableModelReference,
    this.observableSemanticId,
    required this.topic,
    required this.timestamp,
    required this.payload,
  })
  {
    final Reference frameSemanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [I40MessageGlobalReferenceKeys.eventMessage],
    );

    const I40MessageType messageType = I40MessageType.event;
    messageId ??= const Uuid().v4();
    final Sender senderObj = Sender(
      identification: sender,
      conversationRole: ConversationRole.emitter,
    );

    final Frame frame = Frame(
      semanticProtocol: frameSemanticProtocol,
      messageType: messageType,
      messageId: messageId!,
      sender: senderObj,
    );

    final List<SubmodelElement> elements = <SubmodelElement>[];

    //source
    final Reference sourceSemanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [
        I40MessageGlobalReferenceKeys.eventMessageSource,
      ],
    );
    final ReferenceElement sourceElement = ReferenceElement(
      idShort: 'source',
      semanticId: sourceSemanticProtocol,
      value: eventModelReference,
    );
    elements.add(sourceElement);

    //source semantic id
    if (eventSemanticId != null) {
      final Reference sourceSemanticIdSemanticProtocol = Reference(
        type: ReferenceTypes.ExternalReference,
        keys: [I40MessageGlobalReferenceKeys.eventMessageSourceSemanticId,],
      );
      final ReferenceElement sourceSemanticIdElement = ReferenceElement(
        idShort: 'sourceSemanticId',
        semanticId: sourceSemanticIdSemanticProtocol,
        value: eventSemanticId!,
      );
      elements.add(sourceSemanticIdElement);
    }

    //observable
    final Reference observableSemanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [
        I40MessageGlobalReferenceKeys.eventMessageObservableReference,
      ],
    );
    final ReferenceElement observableElement = ReferenceElement(
      idShort: 'observableReference',
      semanticId: observableSemanticProtocol,
      value: observableModelReference,
    );
    elements.add(observableElement);

    //observable semantic id
    if (observableSemanticId != null) {
      final Reference observableSemanticIdSemanticProtocol = Reference(
        type: ReferenceTypes.ExternalReference,
        keys: [I40MessageGlobalReferenceKeys.eventMessageObservableSemanticId,],
      );
      final ReferenceElement observableSemanticIdElement = ReferenceElement(
        idShort: 'observableSemanticId',
        semanticId: observableSemanticIdSemanticProtocol,
        value: observableSemanticId!,
      );
      elements.add(observableSemanticIdElement);
    }

    //topic
    final Reference topicSemanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [
        I40MessageGlobalReferenceKeys.eventMessageTopic,
      ],
    );
    final Property<String> topicElement = Property<String>(
      idShort: 'topic',
      semanticId: topicSemanticProtocol,
      value: topic,
    );
    elements.add(topicElement);

    //timestamp
    final Reference timestampSemanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [
        I40MessageGlobalReferenceKeys.eventMessageTimestamp,
      ],
    );
    final Property<DateTime> timestampElement = Property<DateTime>(
      idShort: 'timestamp',
      semanticId: timestampSemanticProtocol,
      value: timestamp,
    );
    elements.add(timestampElement);

    // payload
    final Reference payloadSemanticProtocol = Reference(
      type: ReferenceTypes.ExternalReference,
      keys: [
        I40MessageGlobalReferenceKeys.eventMessagePayload,
      ],
    );
    final SubmodelElementCollection payloadElement = SubmodelElementCollection(
      idShort: 'payload',
      semanticId: payloadSemanticProtocol,
    )..addElement(payload);

    elements.add(payloadElement);

    super.frame = frame;
    super.interactionElements = elements;
  }

  factory I40EventMessage.fromJson(Map<String, dynamic> json) {
    String sender;
    Reference eventModelReference;
    Reference? eventSemanticId;
    Reference observableModelReference;
    Reference? observableSemanticId;
    String? topic;
    DateTime? timestamp;
    SubmodelElement? payload;

    try {
      final Frame frame = json.containsKey(I40MessageKeys.frame)
          ? Frame.fromJson(json[I40MessageKeys.frame])
          : throw JsonMissingKeyException(I40MessageKeys.frame, json.toString());

      if (!frame.semanticProtocol.keys.contains(I40MessageGlobalReferenceKeys.eventMessage)) {
        throw InvalidMesssageJsonValue(
            'Invalid semantic reference for EventMessage', json.toString());
      }

      sender = frame.sender.identification;

      final List<SubmodelElement> interactionElements = json.containsKey(I40MessageKeys.interactionElements)
          ? createSubmodelElementListFromJson(json[I40MessageKeys.interactionElements] as List<dynamic>)
          : throw JsonMissingKeyException(I40MessageKeys.interactionElements, json.toString());

      ReferenceElement? sourceElement;
      ReferenceElement? sourceSemanticIdElement;
      ReferenceElement? observableElement;
      ReferenceElement? observableSemanticIdElement;
      Property? topicElement;
      Property? timestampElement;
      SubmodelElementCollection? payloadElement;
      for (final SubmodelElement element in interactionElements) {
        if(element is ReferenceElement && element.idShort == 'source') {
          sourceElement = element;
        } else if(element is ReferenceElement && element.idShort == 'sourceSemanticId') {
          sourceSemanticIdElement = element;
        } else if(element is ReferenceElement && element.idShort == 'observableReference') {
          observableElement = element;
        } else if(element is ReferenceElement && element.idShort == 'observableSemanticId') {
          observableSemanticIdElement = element;
        } else if(element is Property && element.idShort == 'topic') {
          topicElement = element;
        } else if(element is Property && element.idShort == 'timestamp') {
          timestampElement = element;
        } else if(element is SubmodelElementCollection && element.idShort == 'payload') {
          payloadElement = element;
        }
      }

      if (payloadElement == null || sourceElement == null || observableElement == null || topicElement == null || timestampElement == null) {
        throw InvalidMesssageJsonValue(
            'No InteractionElement with one of the following shortIds provided [payload, source, observable, topic, timestamp]',
            json[I40MessageKeys.interactionElements].toString());
      }

      eventModelReference = sourceElement.value;

      if(sourceSemanticIdElement != null) {
        eventSemanticId = sourceSemanticIdElement.value;
      }

      observableModelReference = observableElement.value;

      if(observableSemanticIdElement != null) {
        observableSemanticId = observableSemanticIdElement.value;
      }

      topic = topicElement.value as String?;
      timestamp = timestampElement.value as DateTime?;

      if(topic == null || timestamp == null) {
        throw InvalidMesssageJsonValue(
            'No Topic or Timestamp provided',
            json[I40MessageKeys.interactionElements].toString());
      }

      // get actual submodel element
      payload = payloadElement.elements.firstOrNull;
      if (payload == null) {
        throw InvalidMesssageJsonValue(
            'No SubmodelElement as part of the payload provided', payloadElement.toJson().toString());
      }

      return I40EventMessage(
        sender: sender,
        messageId: frame.messageId,
        eventModelReference: eventModelReference,
        eventSemanticId: eventSemanticId,
        observableModelReference: observableModelReference,
        observableSemanticId: observableSemanticId,
        topic: topic,
        timestamp: timestamp,
        payload: payload,
      );
    } on TypeError catch (e) {
      throw InvalidJsonSchemaException(e.stackTrace.toString(), json.toString());
    }
  }
}
