import 'dart:ffi';
import 'dart:io';

import 'package:flutter/services.dart';
import 'package:s3i_flutter/s3i_flutter.dart';
import 'package:s3i_flutter/src/directory/link.dart';
import 'package:s3i_flutter/src/directory/property.dart';
import 'package:s3i_flutter/src/directory/value.dart';
import 'package:s3i_flutter/src/exceptions/json_missing_key_exception.dart';
import 'package:s3i_flutter/src/json_serializable_object.dart';
import 'package:s3i_flutter/src/utils/json_keys.dart';

import 'endpoint.dart';

// TODO(Bek): This was added (do we need it?)

/// Represents a 'Service' in the S3I-Directory data model.

class Service implements JsonSerializableObject {
  /// Creates a [Service] with the given [serviceType][endPoint][parameterTypes][resultTypes].
  Service();

  /// Creates a [Service] from a decoded [json] entry.
  ///
  /// Could throw a [JsonMissingKeyException] if [classString] (or other
  /// required keys) could not be found in the json. Throws a [TypeError]
  /// if the values couldn't be parsed as expected.
  factory Service.fromJson(Map<String, dynamic> json) {

    final Service s = Service()

    ..serviceType = json.containsKey(DirectoryKeys.serveiceType)
        ? json[DirectoryKeys.serveiceType] as String
        : throw JsonMissingKeyException(
        DirectoryKeys.serveiceType, json.toString())
    ..endPoints = json.containsKey(DirectoryKeys.endpoints) ? _createEndpointList(
        json[DirectoryKeys.endpoints] as List<dynamic>)
        : throw JsonMissingKeyException(
        DirectoryKeys.endpoints, json.toString())
    ..parameterTypes = json.containsKey(DirectoryKeys.parameterTypes)? _createParameterTypeList(json[DirectoryKeys.parameterTypes] as Map<String,dynamic>)
        : throw JsonMissingKeyException(DirectoryKeys.parameterTypes, json.toString())
    ..resultTypes = json.containsKey(DirectoryKeys.resultTypes)? _createResultTypeList(json[DirectoryKeys.resultTypes]as dynamic)
          : throw JsonMissingKeyException(DirectoryKeys.resultTypes, json.toString());

    return s;
  }


  late String serviceType;
  late List<Endpoint> endPoints;
  late List<ParameterProperty> parameterTypes;
  late List<ResultProperty> resultTypes;


  /// Maps the [jsonList] to a [List<ParameterType>].
  ///
  /// Throws a [TypeError] if a element in the list could not be parsed. Throws a
  /// [JsonMissingKeyException] if a needed key is missing.
  static List<ParameterProperty> _createParameterTypeList(Map<String,dynamic> jsonMap) {
    return jsonMap.entries.map((e) => ParameterProperty(e.key, e.value.toString())).toList();
  }

  /// Maps the [jsonList] to a [List<ResultType>].
  ///
  /// Throws a [TypeError] if a element in the list could not be parsed. Throws a
  /// [JsonMissingKeyException] if a needed key is missing.
  static List<ResultProperty> _createResultTypeList(dynamic resultsTypes) {
    // if resultTypes is a String, we only have one resultType without corresponding key
    if(resultsTypes is String) {
      return [ResultProperty("result", resultsTypes as String)];
    }
    // else a Map is passed
    else {
      //print(resultsTypes);
      return (resultsTypes as Map<String, dynamic>).entries.map((e) =>
          ResultProperty(e.key, e.value.toString())).toList();
    }
  }

  static List<Endpoint> _createEndpointList(List<dynamic> jsonList) {
    return jsonList.map((dynamic endP) => Endpoint(endP as String)).toList();
  }



  @override
  String toString() {
    return 'Object($serviceType ---- $endPoints ---- $parameterTypes ---- $resultTypes';
  }

  @override
  Map<String, dynamic> toJson() {
    final Map<String, dynamic> newJson = <String, dynamic>{};
    newJson[DirectoryKeys.serveiceType] = serviceType;

    newJson[DirectoryKeys.endpoints] =
          endPoints.map((Endpoint e) => e.endpoint).toList();

    Map paramTypes = <String,String>{};
    parameterTypes.forEach((element) {paramTypes[element.name] = element.type;});
    newJson[DirectoryKeys.parameterTypes] = paramTypes;

    Map resTypes = <String,String>{};
    resultTypes.forEach((element) {resTypes[element.name] = element.type;});
    newJson[DirectoryKeys.resultTypes] = resTypes;
    return newJson;
  }

}


class IoParameterType {

  IoParameterType(this.name, this.type);
  final String name;
  final String type;
}

class ParameterProperty extends IoParameterType{
  ParameterProperty(String name, String type): super(name, type);
}

class ResultProperty extends IoParameterType{
  ResultProperty(String name, String type): super(name, type);
}

