import 'package:dart_aas/dart_aas.dart';
import 'package:s3i_flutter/src/directory/endpoint.dart';
import 'package:s3i_flutter/src/exceptions/invalid_json_schema_exception.dart';
import 'package:s3i_flutter/src/exceptions/json_missing_key_exception.dart';
import 'package:s3i_flutter/src/utils/json_keys.dart';

// TODO(poq): add ditto specific attributes like definition,_revision,_modified

/// Represents a `Thing` in the S3I-Directory data model.
///
/// See the [KWH Standpunkt](https://www.kwh40.de/wp-content/uploads/2020/04/KWH40-Standpunkt-S3I-v2.0.pdf)
/// for more information about a thing an its function.
///
/// Simple UML diagram for the S3I-Directory data model:
/// ```
/// +--------------+
/// |              |
/// |  S3I::Thing  |
/// |              |
/// +--------------+
///         ^
///         |
///         |thingStructure
///         |
///         |0..1
/// +-------+-------+0..1   target    +--------------+
/// |               +---------------->|              |
/// |  S3I::Object  |                 |  S3I::Link   |
/// |               |<----------------+              |
/// +---------------+  links        * +--------------+
///         ^
///         |
///         |values
///         |
///         |*
/// +-------+------+
/// |              |
/// |  S3I::Value  |
/// |              |
/// +--------------+
/// ```
class Thing extends AssetAdministrationShell {
  /// Creates a new [Thing] with the given [endpoints].
  Thing({required super.id, required super.assetInformation, super.idShort,
    super.submodelList});

  /// Creates a [Thing] from a decoded [json] entry.
  factory Thing.fromJson(Map<String, dynamic> map) {
    try {
      final String tId = map.containsKey(DittoKeys.thingId)
          ? map[DittoKeys.thingId] as String
          : throw JsonMissingKeyException(DittoKeys.thingId, map.toString());
      final AssetAdministrationShell? aas = getAASfromJson(map).firstOrNull;
      if (aas == null) {
        throw InvalidJsonSchemaException('No AAS found in json', map.toString());
      }
      final Thing internalThing = Thing(
          id: tId,
          assetInformation: aas.assetInformation,
          idShort: aas.idShort,
          submodelList: aas.submodels);
      if (map.containsKey(DittoKeys.attributes)) {
        final Map<String, dynamic> attributesMap =
            map[DittoKeys.attributes] as Map<String, dynamic>;
        internalThing
          ..allEndpoints = attributesMap.containsKey(DirectoryKeys.allEndpoints)
              ? _createEndpointList(
                  attributesMap[DirectoryKeys.allEndpoints] as List<dynamic>)
              : null
          ..defaultEndpoint = attributesMap
                  .containsKey(DirectoryKeys.defaultEndpoint)
              ? Endpoint(attributesMap[DirectoryKeys.defaultEndpoint] as String)
              : null;
      }
      return internalThing;
    } on TypeError catch (e) {
      throw InvalidJsonSchemaException(e.stackTrace.toString(), map.toString());
    }
  }

  /// All endpoints of this thing (e.g. S3I-B, OPC UA, MQTT, etc - addresses).
  List<Endpoint>? allEndpoints;

  /// The currently active endpoint of the thing.
  Endpoint? defaultEndpoint;

  @override
  Map<String, dynamic> toJson() {
    final Map<String, dynamic> newJson = super.toJson();
    final Map<String, dynamic> attributesMap = <String, dynamic>{};
    newJson[DittoKeys.thingId] = id;
    if (allEndpoints != null) {
      attributesMap[DirectoryKeys.allEndpoints] =
          allEndpoints!.map((Endpoint e) => e.endpoint).toList();
    }
    if (defaultEndpoint != null) {
      attributesMap[DirectoryKeys.defaultEndpoint] = defaultEndpoint!.endpoint;
    }
    if (attributesMap.isNotEmpty) newJson[DittoKeys.attributes] = attributesMap;
    return newJson;
  }

  static List<Endpoint> _createEndpointList(List<dynamic> jsonList) {
    return jsonList.map((dynamic endP) => Endpoint(endP as String)).toList();
  }

}
